#----------------------------------------------------------------------------
# PROCEDURES
#----------------------------------------------------------------------------

# Map C types to the C expression that should be used to initialise them.
# Only the types listed here can be used as array element types.
array set init_values {
    uint8_t {rand ()}
    uint16_t {rand ()}
    uint32_t {rand ()}
    int8_t {rand ()}
    int16_t {rand ()}
    int32_t {rand ()}
    float {((rand() & 0x1ff) - 0x100) * 0x1.0p-4f}
}

# Types which have a special comparison function
array set equal_fns {
    float floats_eq_p
}

# Return the value of test parameter NAME, as given in the spec file.
# The first optional argument is the default value.
proc param { name args } {
    global contents specname

    if { [info exists contents($name)] } {
	return $contents($name)
    }
    if { [llength $args] == 0 } {
	puts stderr "$specname: missing parameter '$name'"
	exit 1
    }
    return [lindex $args 0]
}

# Execute BODY with LINE set to every nonempty line in TEXT.
proc foreach_nonempty_line { line text body } {
    upvar $line up_line

    foreach up_line [split $text "\n"] {
	set up_line [string trim $up_line]
	if { $up_line != "" } {
	    uplevel 1 $body
	}
    }
}

# Interpret the given string as a list of arrays.  Return a list
# of TYPE NAME DIMS START tuples, where TYPE is the type, NAME is the
# array name, DIMS is a list of dimensions, and START is starting index
# for the outermost dimension.
proc parse_arrays { text } {
    global specname

    set arrays [list]
    foreach_nonempty_line line $text {
	set dims [list]
	set rest $line
	if { [regexp {^(.*):\s*(.+)\s*;$} $rest dummy rest start] } {
	    append rest ";"
	} else {
	    set start 0
	}
	if { [regexp {^(\w+)\s+(\w+)\s*(.*);$} $rest dummy type name rest] } {
	    while { $rest != ""
		    && [regexp {^\[([^\[]*)\]\s*(.*)} $rest dummy dim rest] } {
		lappend dims $dim
	    }
	}
	if { $rest != "" } {
	    puts stderr "$specname: unrecognised array definition '$line'"
	    exit 1
	}
	lappend arrays $type $name $dims $start
    }
    return $arrays
}

# Interpret the given string as a list of scalar initialisers.  Return
# a list of TYPE NAME VALUE triplets, where TYPE is the type, NAME is the
# variable name, and VALUE is the initial value.
proc parse_scalars { text } {
    global specname

    set scalars [list]
    foreach_nonempty_line line $text {
	if { ![regexp {^(\w+(?:\s*\*)*)\s*(\w+)\s*=\s*(.*);$} $line \
		   dummy type name value] } {
	    puts stderr "$specname: unrecognised scalar definition '$line'"
	    exit 1
	}
	lappend scalars $type $name $value
    }
    return $scalars
}

# Compare two values A and B of type TYPE and evalute to true if they are
# different.
proc ne_p { type a b } {
    global equal_fns

    if { [info exists equal_fns($type)] } {
	return "!$equal_fns($type) ($a, $b)"
    } else {
	return "$a != $b"
    }
}

# Execute a "foreach VARS LIST BODY" loop in which the body expands
# the string in ARGS.  Treat each string as a separate line of C source
# code and return the block of code that the foreach loop produces.
proc text_foreach { vars list args } {
    foreach var $vars {
	upvar $var up_$var
	lappend upvars up_$var
    }
    set text [list]
    foreach $upvars $list {
	foreach arg $args {
	    lappend text [uplevel 1 [list subst $arg]]
	}
    }
    return [join $text "\n"]
}

# Declare that NAME is an array with element type TYPE and dimensions DIMS.
proc array_decl { type name dims } {
    set line "$type $name"
    foreach dim $dims {
	append line "\[[string trim $dim]\]"
    }
    return "$line;"
}

# Declare that NAME points an array with element type TYPE and dimensions DIMS.
proc pointer_decl { type name dims } {
    set num_dims [llength $dims]
    if { $num_dims == 0 } {
	return "$type $name"
    } elseif { $num_dims == 1 } {
	return "$type *$name"
    } else {
	set text "$type (*$name)"
	foreach dim [lrange $dims 1 end] {
	    append text "\[[string trim $dim]\]"
	}
	return $text
    }
}

# Emit loop headers to iterate over array dimensions DIMS.  Set TEXT
# to a list of for loops, INDENT to the indentation of the loop body,
# and SUFFIX to the series of index operations for the current element.
proc array_loops { dims text indent suffix } {
    upvar text out_text indent out_indent suffix out_suffix

    set index 0
    set out_text [list]
    set out_suffix ""
    set out_indent "  "
    foreach dim $dims {
	set var "i$index"
	incr index

	set dim [string trim $dim]
	lappend out_text \
	    "${out_indent}for (size_t $var = 0; $var < $dim; $var++)"
	append out_suffix "\[$var\]"
	set out_indent "$out_indent  "
    }
}

# Initialize array NAME with random data.  DIMS gives the dimensions of
# the array and TYPE is the element type.
proc array_init { type name dims } {
    global init_values specname

    if { ![info exists init_values($type)] } {
	puts stderr "$specname: no random initialiser defined for type '$type'"
	exit 1
    }
    array_loops $dims text indent suffix
    lappend text "$indent$name$suffix = $init_values($type);"
    return [join $text "\n"]
}

# Execute CODE if arrays A and B are different.  DIMS gives the dimensions of
# the arrays and TYPE is their element type.
proc array_compare { type a b dims code } {
    global init_values specname

    array_loops $dims text indent suffix
    lappend text "${indent}if ([ne_p $type $a$suffix $b$suffix])"
    lappend text "${indent}  $code"
    return [join $text "\n"]
}

# Expand TEXT.  If it is different from the contents of file NAME,
# replace NAME with the new text.
proc genfile { name text } {
    global basename specname

    set filename "autosrc/$basename/$name"
    set contents [uplevel 1 [list subst $text]]

    # Remove leading and trailing blank lines.
    regsub {^\n*} $contents "" contents
    regsub {\n*$} $contents "\n" contents

    # Make sure there is at most one blank line between blocks of text.
    regsub -all {\n\n\n+} $contents "\n\n" contents

    # Make sure that code blocks don't start or end with a blank line.
    regsub -all {\{\n\n+} $contents "{\n" contents
    regsub -all {\n+\n\}} $contents "\n}" contents

    set contents "/* Automatically generated from $specname.  */\n\n$contents"
    if { [catch { open $filename r } file]
	 || ![string equal [read $file] $contents] } {
	set file [open $filename w]
	puts -nonewline $file $contents
	close $file
    }
}

#----------------------------------------------------------------------------
# MAIN SCRIPT
#----------------------------------------------------------------------------

# Parse the arguments.
if { [llength $argv] != 1 } {
    puts stderr "Usage: <spec name>"
    exit 1
}
set basename [lindex $argv 0]

# Read the spec file into an array.
set specname "spec/$basename.txt"
if { [catch { open $specname r } file] } {
    puts stderr $file
    exit 1
}
array set contents [read $file]
close $file

# Create any parts of the output directory that don't already exist.
file mkdir "autosrc/$basename"

# Read and parse spec elements.
set arrays [parse_arrays [param arrays {}]]
set inputs [parse_scalars [param inputs {}]]
set outputs [parse_scalars [param outputs {}]]
set scalars [concat $inputs $outputs]

#----------------------------------------------------------------------------

genfile test.h {
[text_foreach include [concat stdint.h stdlib.h [param includes {}]] \
     {#include <$include>}]

#define COUNT [param count]

[text_foreach { type name } [array get equal_fns] \
     {extern int $name ($type, $type);}]

[text_foreach group { base peak } \
     {[text_foreach { type name dims start } $arrays \
	   {extern [array_decl $type ${group}_$name $dims]}]} \
     {[text_foreach { type name value } $outputs \
	   {extern $type ${group}_$name;}]} \
     ""]
extern void initialise (void);
extern void base_loop (void);
extern void peak_loop (void);
extern void preload (void);
extern const char *compare_loops (void);

extern int base_repeat_count;

[param decls {}]
}

#----------------------------------------------------------------------------

genfile compare.c {
#include "test.h"
#include <string.h>

[text_foreach group { base peak } \
     {[text_foreach { type name dims start } $arrays \
	   {[array_decl $type ${group}_$name $dims] __attribute__((aligned(32)))}]} \
     {[text_foreach { type name value } $outputs \
	   {$type ${group}_$name;}]} \
     ""]
int base_repeat_count = [param repeat];

void __attribute__((noinline))
initialise (void)
{
[text_foreach { type name dims start } $arrays \
     {[array_init $type base_$name $dims]} \
     {  memcpy (peak_$name, base_$name, sizeof (base_$name));}]
}

void
preload (void)
{
[text_foreach { type name dims start } $arrays \
     {  for (size_t i = 0; i < sizeof (peak_$name); i += 32)} \
     {    __asm volatile ("pld \[%0\]" :: "r" ((uintptr_t) peak_$name + i));}]
}

const char *__attribute__((noinline))
compare_loops (void)
{
[text_foreach { type name dims start } $arrays \
     {[array_compare $type base_$name peak_$name $dims "return \"$name\";"]}]
[text_foreach { type name value } $outputs \
     {  if ([ne_p $type base_$name peak_$name])} \
     {    return "$name"; }]
  return 0;
}
}

#----------------------------------------------------------------------------

genfile base.c {
#include "test.h"

void __attribute__((noinline))
base_loop (void)
{
[text_foreach { type name dims start } $arrays \
     {  [pointer_decl $type $name $dims] = &base_$name\[$start\];}]
[text_foreach { type name value } $scalars \
     {  $type $name = $value;}]

[param loop]

[text_foreach { type name value } $outputs \
     {  base_$name = $name;}]
}
}

#----------------------------------------------------------------------------

genfile peak.c {
#include "test.h"

[text_foreach { type name dims start } $arrays \
     {static [pointer_decl $type "volatile vol_$name" $dims] = &peak_$name\[$start\];}]

[text_foreach { type name dims start } $arrays \
     {#define $name peak_$name}]

[text_foreach { type name value } $inputs \
     {static $type volatile vol_$name = $value;}]

[text_foreach { type name dims start } $arrays \
     {#undef $name}]

void __attribute__((noinline))
peak_loop (void)
{
[text_foreach { type name dims start } $arrays \
     {  [pointer_decl $type "__restrict $name" $dims] = vol_$name;}]
[text_foreach { type name value } $inputs \
     {  $type $name = vol_$name;}]

[text_foreach { type name value } $outputs \
     {  $type $name = $value;}]
[param loop]

[text_foreach { type name value } $outputs \
     {  peak_$name = $name;}]
}
}
